{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Listas Enlazadas\n",
    "\n",
    "Recuerden que estas estructuras se suelen colocar en una clase mayor de Java llamada `Collection`\n",
    "\n",
    "```{figure} ../../images/Figure20.1.png\n",
    "---\n",
    "width: 95%\n",
    "name: Figure20.1\n",
    "---\n",
    "Here is my figure caption!\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Listas Simples Enlazadas con Nodos\n",
    "\n",
    "Una **lista simple enlazada** es una estructura de datos lineal formada por nodos, donde cada nodo tiene un dato y una referencia al siguiente nodo de la lista.\n",
    "\n",
    "```{figure} ../../images/Figure2.16.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure2.16\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "```{figure} ../../images/Figure24.7.png\n",
    "---\n",
    "width: 80%\n",
    "name: Figure24.7\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "Hay una clase de java ya implementada llamada [LinkedList](https://github.com/openjdk/jdk/blob/master/src/java.base/share/classes/java/util/LinkedList.java).\n",
    "\n",
    "```{figure} ../../images/Figure24.11.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure24.11\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "### Implementación en Java"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "public class Nodo {\n",
    "  int data; // Dato almacenado en el nodo\n",
    "  Nodo siguiente; // Referencia al siguiente nodo\n",
    "\n",
    "  public Nodo(int data) {\n",
    "      this.data = data;\n",
    "      this.siguiente = null; // Inicialmente, no tiene un nodo siguiente\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "public class ListaSimpleEnlazada {\n",
    "  private Nodo cabeza; // Primer nodo de la lista\n",
    "\n",
    "  public ListaSimpleEnlazada() {\n",
    "      cabeza = null; // La lista comienza vacía\n",
    "  }\n",
    "\n",
    "  // Método para agregar un nodo al final de la lista\n",
    "  public void agregar(int data) {\n",
    "      Nodo nuevoNodo = new Nodo(data);\n",
    "      if (cabeza == null) {\n",
    "        cabeza = nuevoNodo; // Si la lista está vacía, el nuevo nodo es la cabeza\n",
    "      } \n",
    "      else {\n",
    "        Nodo temp = cabeza;\n",
    "        while (temp.siguiente != null) {\n",
    "            temp = temp.siguiente; // Avanza al último nodo\n",
    "        }\n",
    "        temp.siguiente = nuevoNodo; // Se enlaza el nuevo nodo al final\n",
    "      }\n",
    "  }\n",
    "\n",
    "  // Método para imprimir la lista\n",
    "  public void imprimir() {\n",
    "      Nodo temp = cabeza;\n",
    "      while (temp != null) {\n",
    "          System.out.print(temp.data + \" -> \");\n",
    "          temp = temp.siguiente;\n",
    "      }\n",
    "      System.out.println(\"null\");\n",
    "  }\n",
    "  \n",
    "  public static void main(String[] args) {\n",
    "      ListaSimpleEnlazada lista = new ListaSimpleEnlazada();\n",
    "      lista.agregar(10);\n",
    "      lista.agregar(20);\n",
    "      lista.agregar(30);\n",
    "      lista.imprimir(); // Imprime: 10 -> 20 -> 30 -> null\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "// probar clase aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Nodos en Listas Doblemente Enlazadas\n",
    "\n",
    "Una **lista doblemente enlazada** es una estructura en la que cada nodo tiene dos referencias:\n",
    "- Una referencia al siguiente nodo.\n",
    "- Una referencia al nodo anterior.\n",
    "\n",
    "Esto permite recorrer la lista en ambas direcciones.\n",
    "\n",
    "```{figure} ../../images/Figure2.19.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure2.19\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n",
    "### Implementación en Java"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "class NodoDoble {\n",
    "  int data;\n",
    "  NodoDoble siguiente;\n",
    "  NodoDoble anterior;\n",
    "\n",
    "  public NodoDoble(int data) {\n",
    "      this.data = data;\n",
    "      this.siguiente = null;\n",
    "      this.anterior = null;\n",
    "  }\n",
    "}\n",
    "\n",
    "public class ListaDobleEnlazada {\n",
    "  private NodoDoble cabeza;\n",
    "\n",
    "  public ListaDobleEnlazada() {\n",
    "      cabeza = null;\n",
    "  }\n",
    "\n",
    "  // Método para agregar un nodo al final de la lista\n",
    "  public void agregar(int data) {\n",
    "      NodoDoble nuevoNodo = new NodoDoble(data);\n",
    "      if (cabeza == null) {\n",
    "          cabeza = nuevoNodo;\n",
    "      } else {\n",
    "          NodoDoble temp = cabeza;\n",
    "          while (temp.siguiente != null) {\n",
    "              temp = temp.siguiente;\n",
    "          }\n",
    "          temp.siguiente = nuevoNodo;\n",
    "          nuevoNodo.anterior = temp; // El nuevo nodo apunta al anterior\n",
    "      }\n",
    "  }\n",
    "\n",
    "  // Método para imprimir la lista de forma directa\n",
    "  public void imprimirDirecto() {\n",
    "      NodoDoble temp = cabeza;\n",
    "      while (temp != null) {\n",
    "          System.out.print(temp.data + \" <-> \");\n",
    "          temp = temp.siguiente;\n",
    "      }\n",
    "      System.out.println(\"null\");\n",
    "  }\n",
    "\n",
    "  // Método para imprimir la lista de forma inversa\n",
    "  public void imprimirInverso() {\n",
    "      if (cabeza == null) return;\n",
    "      NodoDoble temp = cabeza;\n",
    "      while (temp.siguiente != null) {\n",
    "          temp = temp.siguiente; // Avanza al último nodo\n",
    "      }\n",
    "      while (temp != null) {\n",
    "          System.out.print(temp.data + \" <-> \");\n",
    "          temp = temp.anterior; // Retrocede en la lista\n",
    "      }\n",
    "      System.out.println(\"null\");\n",
    "  }\n",
    "\n",
    "  public static void main(String[] args) {\n",
    "      ListaDobleEnlazada lista = new ListaDobleEnlazada();\n",
    "      lista.agregar(10);\n",
    "      lista.agregar(20);\n",
    "      lista.agregar(30);\n",
    "      lista.imprimirDirecto();  // Imprime: 10 <-> 20 <-> 30 <-> null\n",
    "      lista.imprimirInverso();  // Imprime: 30 <-> 20 <-> 10 <-> null\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "// probar clases y métodos aquí"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Nodos en Grafos\n",
    "\n",
    "Un **grafo** es una estructura de datos que consiste en nodos (llamados vértices) y enlaces entre ellos (aristas). Los grafos pueden ser dirigidos o no dirigidos, dependiendo de si las conexiones entre los nodos tienen una dirección.\n",
    "\n",
    "En un grafo:\n",
    "- Los **nodos** representan entidades.\n",
    "- Las **aristas** representan las relaciones entre estas entidades.\n",
    "\n",
    "### Implementación básica de un grafo usando nodos\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "import java.util.ArrayList;\n",
    "import java.util.List;\n",
    "\n",
    "class NodoGrafo {\n",
    "    int data;\n",
    "    List<NodoGrafo> vecinos; // Lista de nodos vecinos\n",
    "\n",
    "    public NodoGrafo(int data) {\n",
    "        this.data = data;\n",
    "        vecinos = new ArrayList<>();\n",
    "    }\n",
    "\n",
    "    // Método para agregar una arista (relación) entre este nodo y otro\n",
    "    public void agregarVecino(NodoGrafo vecino) {\n",
    "        vecinos.add(vecino);\n",
    "    }\n",
    "}\n",
    "\n",
    "public class Grafo {\n",
    "    private List<NodoGrafo> nodos;\n",
    "\n",
    "    public Grafo() {\n",
    "        nodos = new ArrayList<>();\n",
    "    }\n",
    "\n",
    "    // Método para agregar un nodo al grafo\n",
    "    public void agregarNodo(int data) {\n",
    "        NodoGrafo nuevoNodo = new NodoGrafo(data);\n",
    "        nodos.add(nuevoNodo);\n",
    "    }\n",
    "\n",
    "    // Método para conectar dos nodos\n",
    "    public void conectarNodos(NodoGrafo nodo1, NodoGrafo nodo2) {\n",
    "        nodo1.agregarVecino(nodo2);\n",
    "        nodo2.agregarVecino(nodo1); // Grafo no dirigido\n",
    "    }\n",
    "\n",
    "    public static void main(String[] args) {\n",
    "        Grafo grafo = new Grafo();\n",
    "        \n",
    "        NodoGrafo nodo1 = new NodoGrafo(1);\n",
    "        NodoGrafo nodo2 = new NodoGrafo(2);\n",
    "        NodoGrafo nodo3 = new NodoGrafo(3);\n",
    "\n",
    "        grafo.conectarNodos(nodo1, nodo2);\n",
    "        grafo.conectarNodos(nodo2, nodo3);\n",
    "\n",
    "        System.out.println(\"El nodo 1 está conectado con: \" + nodo1.vecinos.size() + \" nodo(s).\");\n",
    "        System.out.println(\"El nodo 2 está conectado con: \" + nodo2.vecinos.size() + \" nodo(s).\");\n",
    "    }\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Enlazadas Circularmente\n",
    "\n",
    "<div align=\"center\">\n",
    "  <img src=\"https://study.com/cimages/multimages/16/circularly_linked_lists.png\" width=40%>\n",
    "</div>\n",
    "\n",
    "```{figure} ../../images/Figure24.18.png\n",
    "---\n",
    "width: 90%\n",
    "name: Figure24.18\n",
    "---\n",
    "Here is my figure caption!\n",
    "```\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Ejercicio\n",
    "\n",
    "::::{admonition} Taller 5\n",
    "\n",
    "Implementen los códigos de nodos, listas simples y doblemente enlazadas de las tres referencias y la clase de LinkedList de java. Este taller se hace de a cuatro personas, cada uno elije un libro diferente y la idea es comparar los códigos.\n",
    "\n",
    "La idea es que estos códigos son implementaciones más completas y robustas de lo que vimos en clase, clases con muchos métodos y una implementación optimizada.\n",
    "\n",
    "No olviden que deben implementar los códigos y probarlos creando objetos y ejecutando sus métodos. Además deben presentar un documento en formato markdown donde se hable de la comparación entre las implementaciones: ventajas, desventajas, métodos, diferencias, etc.\n",
    "\n",
    "::::"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Conclusiones\n",
    "\n",
    "Los nodos y las referencias son componentes fundamentales en la implementación de estructuras de datos como listas enlazadas, listas doblemente enlazadas y grafos. Al dominar estos conceptos, los estudiantes pueden implementar estructuras de datos más complejas y optimizadas, cruciales para la resolución eficiente de problemas en la programación. Las listas permiten manejar datos dinámicos, mientras que los grafos ofrecen una manera poderosa de modelar relaciones entre entidades.\n",
    "\n",
    "\n",
    "## Recursos Adicionales\n",
    "\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Java",
   "language": "java",
   "name": "java"
  },
  "language_info": {
   "codemirror_mode": "java",
   "file_extension": ".jshell",
   "mimetype": "text/x-java-source",
   "name": "java",
   "pygments_lexer": "java",
   "version": "21.0.4+7-Ubuntu-1ubuntu224.04"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
